﻿using System;
using System.Collections;
using System.Collections.Generic;

namespace HeuristicLab.Algorithms.DataAnalysis.SymRegGrammarEnumeration.GrammarEnumeration {

  public class StoredSymbolString {
    public readonly int Hash;
    public readonly SymbolString SymbolString;

    public StoredSymbolString(int hash, SymbolString symbolString) {
      Hash = hash;
      SymbolString = symbolString;
    }
  }

  public enum StorageType {
    PriorityQueue, Stack, Queue, RandomList
  }

  class SearchDataStore : IEnumerable<SymbolString> {

    private Dictionary<int, SymbolString> storedValues; // Store hash-references and associated, actual values 
    private Action<int, double> storeInternal; // Store hash-references 
    private Func<int> fetchInternal; // Fetch hash-reference

    public SearchDataStore(StorageType storageType) {
      storedValues = new Dictionary<int, SymbolString>();

      switch (storageType) {
        case StorageType.PriorityQueue:
          InitPriorityQueue();
          break;
        case StorageType.Stack:
          InitStack();
          break;
        case StorageType.Queue:
          InitQueue();
          break;
        case StorageType.RandomList:
          InitRandomList();
          break;
      }

    }

    #region SearchStrategies

    private void InitPriorityQueue() {
      PriorityQueue<double, int> queue = new PriorityQueue<double, int>(double.MaxValue, double.MinValue, (int)Math.Pow(2.0, 20.0) / 4);
      storeInternal = (hash, prio) => queue.Insert(prio, hash);
      fetchInternal = () => {
        int ret = queue.PeekMinValue();
        queue.RemoveMin();
        return ret;
      };
    }

    private void InitStack() {
      Stack<int> stack = new Stack<int>();

      storeInternal = (hash, prio) => stack.Push(hash);
      fetchInternal = () => stack.Pop();
    }

    private void InitQueue() {
      Queue<int> queue = new Queue<int>();

      storeInternal = (hash, prio) => queue.Enqueue(hash);
      fetchInternal = () => queue.Dequeue();
    }

    private void InitRandomList() {
      List<int> list = new List<int>();
      System.Random rand = new System.Random(999);

      storeInternal = (hash, prio) => list.Add(hash);
      fetchInternal = () => {
        int indexOfHash = rand.Next(list.Count);
        int result = list[indexOfHash];
        list.RemoveAt(indexOfHash);  // TODO: beware this is O(n), at some point in time we should fix this. Maybe change to priority queue with random key.
        return result;
      };
    }

    #endregion

    #region Interface

    public StoredSymbolString GetNext() {
      int hash = fetchInternal.Invoke();
      SymbolString result = storedValues[hash];
      storedValues.Remove(hash);
      return new StoredSymbolString(hash, result);
    }

    public void Store(int hash, double priority, SymbolString s) {
      storeInternal(hash, priority);
      storedValues[hash] = s;
    }

    public bool Contains(int hash) {
      return storedValues.ContainsKey(hash);
    }

    #endregion

    #region Collection-Interface

    public int Count {
      get { return storedValues.Count; }
    }

    public IEnumerator<SymbolString> GetEnumerator() {
      return storedValues.Values.GetEnumerator();
    }

    IEnumerator IEnumerable.GetEnumerator() {
      return GetEnumerator();
    }

    #endregion
  }
}
